from typing import Any
from .module import Module

class Flatten(Module):
    __constants__: Any = ...
    start_dim: Any = ...
    end_dim: Any = ...
    def __init__(self, start_dim: int = ..., end_dim: int = ...) -> None: ...
    def forward(self, input: Any): ...  # type: ignore
    def __call__(self, input: Any): ...  # type: ignore
